/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.interceptor.description;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.xml.transform.TransformerException;

import org.objectweb.petals.component.framework.ExtensionsConstants;
import org.objectweb.petals.component.framework.PEtALSCFException;
import org.objectweb.petals.component.framework.interceptor.Interceptor;
import org.objectweb.petals.component.framework.util.Extensions;
import org.objectweb.petals.component.framework.util.XMLUtil;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;
import org.objectweb.petals.jbi.descriptor.JBIDescriptorBuilder;
import org.objectweb.petals.jbi.descriptor.JBIDescriptorException;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Christophe HAMERLING (chamerling) - eBM WebSourcing
 *
 */
public class InterceptorBuilder {
    private static InterceptorBuilder INSTANCE = null;

    public static final String DELIM = ",";

    public static final String ACCEPT = "accept";

    public static final String SEND = "send";

    public static final String INTERCEPTOR_NODE = "interceptor";

    public static final String PARAMETER_NODE = "param";

    public static final String ORDER_NODE = "order";

    public static final String CLASS_ATTR = "class";

    public static final String ACTIVE_ATTR = "active";

    public static final String MODE_ATTR = "mode";

    public static final String NAME_ATTR = "name";

    public static final String VALUE_ATTR = "value";

    /**
     * Get loader instance
     *
     * @return
     */
    public synchronized static final InterceptorBuilder getInstance() {
        if (INSTANCE == null) {
            INSTANCE = new InterceptorBuilder();
        }
        return INSTANCE;
    }

    /**
     * Creates a new instance of {@link InterceptorBuilder}
     *
     */
    private InterceptorBuilder() {
    }

    /**
     * Load the component interceptors.
     *
     * @param rootPath
     */
    public List<InterceptorConfiguration> loadComponentInterceptors(
            String rootPath) throws PEtALSCFException {
        File jbiFile = new File(rootPath,
                JBIDescriptorBuilder.JBI_DESCRIPTOR_RESOURCE);
        JBIDescriptor descriptor = null;

        try {
            JBIDescriptorBuilder builder = JBIDescriptorBuilder.getInstance();
            descriptor = builder.buildDescriptor(jbiFile.toURI());
        } catch (JBIDescriptorException e) {
            throw new PEtALSCFException("Can not load the JBI descriptor file");
        }

        org.objectweb.petals.jbi.descriptor.Extensions descExtensions = descriptor
                .getComponent().getExtensions();
        if (descExtensions == null) {
            return null;
        }
        Extensions extensions = new Extensions(descExtensions);

        return this.loadInterceptors(extensions);
    }

    /**
     *
     * @param extensions
     * @return
     */
    public List<InterceptorConfiguration> loadInterceptors(Extensions extensions) {
        List<InterceptorConfiguration> interceptors = new ArrayList<InterceptorConfiguration>();

        Node interceptorNode = XMLUtil.findChild(extensions
                .getDocumentFragment(), ExtensionsConstants.INTERCEPTORS, true);
        if (interceptorNode != null) {
            List<Node> children = XMLUtil.getNodeChildren(interceptorNode);
            for (Node childNode : children) {
                InterceptorConfiguration conf = this
                        .loadInterceptorDescriptor(childNode);
                if (conf != null) {
                    interceptors.add(conf);
                }
            }
        }
        return interceptors;
    }

    /**
     * Load the interceptor configuration from the extensions
     *
     * @param extensions
     * @return interceptor configuration defined in the SU or null if no
     *         interceptor extension has been defined in the extensions.
     */
    public ExtensibleElementInterceptorConfiguration loadSUInterceptor(
            Extensions extensions) {
        ExtensibleElementInterceptorConfiguration result = new ExtensibleElementInterceptorConfiguration();

        if (extensions != null && extensions.getDocumentFragment() != null) {
            Node interceptorsNode = XMLUtil.findChild(extensions
                    .getDocumentFragment(), ExtensionsConstants.INTERCEPTORS,
                    true);

            if (interceptorsNode != null) {

                Node inNode = XMLUtil.findChild(interceptorsNode, ACCEPT, true);
                Node outNode = XMLUtil.findChild(interceptorsNode, SEND, true);

                result
                        .setInputInterceptors(this
                                .getOrderedInterceptors(inNode));
                result.setOutputInterceptors(this
                        .getOrderedInterceptors(outNode));
            }
        }
        return result;
    }

    /**
     *
     * @param orderNode
     * @return
     */
    private Map<String, InterceptorConfiguration> getOrderedInterceptors(
            Node orderNode) {
        Map<String, InterceptorConfiguration> result = new LinkedHashMap<String, InterceptorConfiguration>();
        if (orderNode != null
                && (orderNode.getNodeName().equals(ACCEPT) || orderNode
                        .getNodeName().equals(SEND))) {

            List<Node> nodes = XMLUtil.getNodeChildren(orderNode);
            for (Node node : nodes) {
                InterceptorConfiguration config = this
                        .loadInterceptorDescriptor(node);
                result.put(config.getName(), config);
            }
        }
        return result;
    }

    /**
     * Load the interceptor descriptor from the given node.
     *
     * @param interceptorNode
     *            the {@link Node} to create interceptor description from.
     * @return an {@link InterceptorConfiguration} or null if it is not an
     *         interceptor node or if the node is null.
     */
    public InterceptorConfiguration loadInterceptorDescriptor(
            final Node interceptorNode) {
        InterceptorConfiguration result = null;

        if (interceptorNode != null
                && interceptorNode.getNodeName().equals(INTERCEPTOR_NODE)) {
            result = new InterceptorConfiguration();
            NamedNodeMap atts = interceptorNode.getAttributes();
            if (atts == null) {
                return result;
            }

            // get the name, class and active attributes
            Node itcNameNode = atts.getNamedItem(NAME_ATTR);
            Node itcClassNode = atts.getNamedItem(CLASS_ATTR);
            Node itcActiveNode = atts.getNamedItem(ACTIVE_ATTR);
            if (itcNameNode != null) {
                result.setName(itcNameNode.getNodeValue());
            }
            if (itcClassNode != null) {
                result.setClazz(itcClassNode.getNodeValue());
            }
            if (itcActiveNode != null
                    && itcActiveNode.getNodeValue().equals("true")) {
                result.setActive(true);
            }

            // get the parameters
            Map<String, String> parameters = new HashMap<String, String>();
            NodeList paramNodes = interceptorNode.getChildNodes();
            for (int i = 0; i < paramNodes.getLength(); i++) {
                Node n = paramNodes.item(i);
                if (n != null && n.getNodeName().equals(PARAMETER_NODE)) {
                    Node attName = n.getAttributes().getNamedItem(NAME_ATTR);
                    if (attName != null) {
                        NodeList list = n.getChildNodes();
                        String value = "";
                        try {
                            value = XMLUtil.createStringFromDOMNodeList(list);
                        } catch (TransformerException e) {
                        }
                        parameters.put(attName.getNodeValue(), value.trim());
                    }
                }
            }
            result.setParameters(parameters);
        }
        return result;
    }

    /**
     * Load an interceptor instance
     *
     * @return
     */
    public Interceptor loadInstance(String className) throws PEtALSCFException {
        Interceptor interceptor = null;
        Class clazz = this.loadClass(className);

        try {
            Object o = clazz.newInstance();
            if (o instanceof Interceptor) {
                interceptor = (Interceptor) o;
            } else {
                throw new PEtALSCFException("Invalid class, " + className
                        + " is not an interceptor");
            }

        } catch (InstantiationException e) {
            throw new PEtALSCFException("Can not instanciate interceptor", e);
        } catch (IllegalAccessException e) {
            throw new PEtALSCFException("Illegal access to interceptor", e);
        }

        return interceptor;
    }

    /**
     *
     * @param classloader
     * @param className
     * @return
     * @throws JBIException
     */
    protected Class loadClass(String className) throws PEtALSCFException {
        Class clazz;
        ClassLoader classloader = this.getClass().getClassLoader();
        try {
            clazz = classloader.loadClass(className);
            if (clazz == null) {
                throw new PEtALSCFException(
                        "Could not find interceptor class : " + className);
            }

        } catch (ClassNotFoundException e) {
            throw new PEtALSCFException("Class not found interceptor class : "
                    + className, e);
        } catch (NullPointerException e) {
            throw new PEtALSCFException("Null class : " + className);
        }
        return clazz;
    }
}
